/* Copyright 1998-2017 Research Foundation State University of New York */

/* This file is part of QUB Online.                                     */

/* QUB Online is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation, either version 3 of the License, or     */
/* (at your option) any later version.                                   */

/* QUB Online is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/* GNU General Public License for more details.                          */

/* You should have received a copy of the GNU General Public License,    */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see */
/* <http://www.gnu.org/licenses/>.                                       */

#include <algorithm>
using std::max;
using std::min;
#include <limits.h>
#include <string.h>
#include "matrix.h"

#include "qubx_skm.h"
#include "asmffi.h"
using std::endl;

#define SKM_LLCONV .01


//-----------------------------------------------------------------------
// q-fq::matrix utility  - Fill in q and dq arrays.
// Note : drug[] could be NULL and nz could be zero
void skm_xtoq(int nstate, int npath, int* path, double* drug, double volt, 
              double* x, double* q, int nz, double* dx, double* dq) {
  int  i,j,k,m;
  double nQ;
  double c;
  
  //-- Zero the q and dq arrays 
  for (i=0; i<nstate; i++) {
    for (j=0; j<nstate; j++) {
      q[i*nstate+j] = 0.0;						// q[i][j]
      for (k=0; k<nz; k++) 
        dq[i*nstate*nz+j*nz+k] = 0.0;			// dq[i][j][k]         
    }
  }
  
  //-- Fill in rates and diagonal = -sum(rates) 
  for (m=0; m<npath; m++) {
    i = path[m*4+0];								// From State
    j = path[m*4+1];								// To State
    c = ((drug==NULL) ? 1.0 : drug[path[m*4+2]]);	// Drug index 
    // Fill in Q 
    nQ = c * exp(x[2*m]+volt*x[2*m+1]);
    q[i*nstate+j] = nQ;
    q[i*nstate+i] -= nQ;
    
    // Fill in dQ - derivatives to the z-variables  
    for (k=0; k<nz; k++)
      dq[i*nstate*nz+i*nz+k] -= 
        ( dq[i*nstate*nz+j*nz+k] =  nQ * ( dx[2*m*nz+k] + volt*dx[(2*m+1)*nz+k] ));  
  }
}

//-----------------------------------------------------------------------
void skm_qtomutiq(int nstate, int nmutistate, 
                  int* mutistate, int nmutipath, int* mutipath, double* q, 
                  double* mutiq, int nz, double* dq, double* dmutiq){   
	// nz could be zero 
	int  I,J,i,j,k,m;
	double  s;
    
	// rate constants 
	for (I=0; I<nmutistate; I++) 
		for (J=0; J<nmutistate; J++) 
			mutiq[I*nmutistate+J] = 0.0;
		
	for (m=0; m<nmutipath; m++) {
		I = mutipath[m*4+0];
		J = mutipath[m*4+1];
		i = mutipath[m*4+2];
		j = mutipath[m*4+3];
		mutiq[I*nmutistate+J] = mutistate[I*(nstate+1)+i+1]*q[i*nstate+j];
		}
	
	for (I=0; I<nmutistate; I++) {
		for (s=0.,J=0; J<nmutistate; J++) 
			if (J != I) 
				s -= mutiq[I*nmutistate+J];
			mutiq[I*nmutistate+I] = s;
		}
    
	// derivatives to the z-variables 
	for (I=0; I<nmutistate; I++) 
		for (J=0; J<nmutistate; J++) 
			for (k=0; k<nz; k++) 
				dmutiq[I*nmutistate*nz+J*nz+k] = 0.0;
			
	for (m=0; m<nmutipath; m++) {
		I = mutipath[m*4+0];
		J = mutipath[m*4+1];
		i = mutipath[m*4+2];
		j = mutipath[m*4+3];
		for (k=0; k<nz; k++) 
			dmutiq[I*nmutistate*nz+J*nz+k] =                // dmutiq[I][J][k]
			  mutistate[I*(nstate+1)+i+1]*dq[i*nstate*nz+j*nz+k];  
		} 
	
	for (k=0; k<nz; k++) {
		for (I=0; I<nmutistate; I++) {
			for (s=0.,J=0; J<nmutistate; J++) 
				if (J != I) 
					s -= dmutiq[I*nmutistate*nz+J*nz+k];      // dmutiq[I][J][k]
				dmutiq[I*nmutistate*nz+I*nz+k]=s;               // dmutiq[I][I][k]
			}
		}
	}

/* -----------------------------------------------------------------------
	Calculate A = I*t^i0 + Qt^(i0+1)/(i0+1) + Q^2*t^(i0+2)/(i0+2) + ... 
	matrix exponential and its derivatives.
*/
void skm_mexp(int n, float t, double* q, double* a, int nz, double* dq, double* da, int i0) {
	// nz could be zero, dq and da could be NULL
	int i,j,k,l;
	double s;
	double qsum=0.0;				// Sum of abs(t*sum(q)) = magnitude of Q
   
	// initialization
	double * u  = new double[n*n];
	double * v  = new double[n*n*nz];
	double * u1 = new double[n*n];
	double * v1 = new double[n*n*nz];
	
	for (i=0; i<n; i++)    
		for (j=0; j<n; j++) {
			a[i*n+j] = 0.0;
			u[i*n+j] = (i==j)?pow(t,i0):0.0;
			for (k=0; k<nz; k++){
				da[i*n*nz+j*nz+k] = 0.0;
				v[i*n*nz+j*nz+k] = 0.0;
				}
			qsum+=fabs(q[i*n+j]);
			}
	qsum*=t;
		
	//----- series summation for minimum of 4 <iterm> terms until convergence 
	double r=1.0;
	for( int iterm=0; iterm<4 || r/(1.0-r/double(iterm+2))>=1.0e-10; iterm++) {
		for (i=0; i<n; i++)         
			for (j=0; j<n; j++) {
				a[i*n+j] += u[i*n+j];                        // a[i][j]
				for (k=0; k<nz; k++) 
					da[i*n*nz+j*nz+k] += v[i*n*nz+j*nz+k];    // da[i][j][k]
				}
		for (i=0; i<n; i++)
			for (j=0; j<n; j++) { 
				for (s=0.,l=0; l<n; l++) 
					s += u[i*n+l]*q[l*n+j];                   // u[i][l], q[l][j]
				u1[i*n+j]=s; 
				} 
		for (k=0; k<nz; k++)  
			for (i=0; i<n; i++)         
				for (j=0; j<n; j++) {
					for (s=0.,l=0; l<n; l++) 
						s += q[i*n+l]*v[l*n*nz+j*nz+k] + dq[i*n*nz+l*nz+k]*u[l*n+j];
					v1[i*n*nz+j*nz+k] = s;                    // v1[i][j][k]
					}
		for (i=0; i<n; i++)         
			for (j=0; j<n; j++) {
			  u[i*n+j] = u1[i*n+j]*t/(double(iterm+1+i0));
				for (k=0; k<nz; k++) 
				  v[i*n*nz+j*nz+k] = v1[i*n*nz+j*nz+k]*t/(double(iterm+1+i0)); // v[i][j][k]
				}
		
		// Convergence calculation pt. 2
		r *= qsum/double(iterm+1);
		}
	
	// free memory
	delete [] u;
	delete [] v;
	delete [] u1;
	delete [] v1;
	}

//-----------------------------------------------------------------------------
// multi-channel Markov model creation : this function will make a multichannel 
// markov model out of a single channel markov model
// Input nchannel, nstate, npath, state, path, ratio
// Output nmutistate, nmutipath, nmuticlass, mutistate, mutipath
//-----------------------------------------------------------------------
void skm_mutimakv(int nchannel, int nstate, int npath, int *state, int *path, double *ratio, 
				  int& nmutistate, int& nmutipath, int& nmuticlass, int*& mutistate, int*& mutipath){
	int i,k,l,m,n,s,t, iPaths=0, iState;
	
    //-- Determine whether open is negative or positive and use to set sort order 
	// for muticlasses in eclass().
	// Note : The try catch is probably not effective here.  Accessing beyond the 
	// bounds of the array would likely run into more data and not throw an exception.
	// Better to ensure that the passed array ratio[] has an element [1] always.
	double openDirection;		
	try {			
		openDirection = ratio[1];
		}
	catch (...) {
		openDirection = 1.0;
		}

	//-- Precalc model size for memory allocation 
	// Calculates the number of Combinations of channels/states per row in mutistate.
	// For each state, add to temp log(state+nchann)-log(state) and then antilog on the sum
	double tmp=0.0;
	for( n=1; n<nstate; n++) 
		tmp += log(double(n+nchannel))-log(double(n));
	n = int(exp(tmp)+1);						//estimates nMutiState - may be 1 greater than necc.
	mutistate = new int[n * (nstate+1)];		//create empty mutistate array.
	float * aCond = new float[n];		// Conductance class sums
	
	// Create a 2d style array for accessing mutistate array 
	int ** aMutiState = new int * [n];
	for( i=0; i<n; ++i ) 
		aMutiState[i] = &(mutistate[i*(nstate+1)]);

	//----- Fill in mutistate , acond, iPaths.
	// mutistate is an integer array of (estimated nMutiState) * (nstate+1)
	// Each (nState+1) group is a 1-based array of number of channels in each state,
	// and [0] element is set to the eclass result class for the row
	// iterate through combinations of states / channels and set nMutiState
	// aCond is the effective conductance <aCond[]> = sum[all states] ( channels * ratio ) 
	// iPaths is the number of mutistate connections which have only one channel transition
	nmutistate=0;
	CChannelCombine oMeta(nchannel, nstate);
	do {
		aCond[nmutistate]=0.0;
		for ( i=0; i<nstate; ++i ) {
			iState=oMeta.State0(i);
			aMutiState[nmutistate][i+1]=iState;
			aCond[nmutistate] += (float) (iState * ratio[state[i]]);
			if( iState > 0 )								// pre calc # of paths 
				iPaths += (nstate-1);
			}
		++nmutistate;
		} while( oMeta.GetNextCombine() );

	//----- conductance classes 
	float * aClasses = new float[nmutistate];	// output buffer, discarded unused
	int *  ia = new int[nmutistate];			// mapping of data to classes
	
	eclass( aCond, nmutistate, nmuticlass, aClasses, ia, openDirection);		// Fill C,ia with classes
	
	for (n=0; n<nmutistate; n++) 
		aMutiState[n][0] = ia[n];
	
	delete [] aCond;
	delete [] aClasses;
	delete [] ia;

	//-----
	// precalc the number of transition paths for memory allocation
	// iterates through all combinations of mutistates and counts those 
	// which differ only by 1 transition.
	// New method : for each non zero state, add paths for the number of 
	//  places one of the channels could go.  This is precalced above.
	// This calculates a maximum possible, but the actual number used is less
	/*
	n=0;
	for(k=0; k<nmutistate; k++)  
		for (l=0; l<nmutistate; l++) { 
			s=0;
			for (i=0; i<nstate; i++)  
				s += int(fabs(aMutiState[k][i+1]-aMutiState[l][i+1])); 
			if (s==2) 
				n++; 
			} 
	mutipath = new int[n*4];
	*/
	mutipath = new int[iPaths*4];
	
	// transition paths - for each mutipath pair (m,n), if they differ 
	// by only 1 channel state change then find the original states (k,l)
	// and check if a path exists between k  & l -->> add a mutipath if so.
	//
	// Note : Consider a transition where  two channels change state in a 
	// single delta T - is this accounted for elsewhere ? {JB}
	// In a simple 2 state 3 channel model currently (C,C,C) we are only 
	// setting mutipaths for (C,C,O),(C,O,C),(O,C,C) but all states 
	// including (O,O,O) are reachable in a single delta T ?
	//
	nmutipath = 0; 
	for( m=0; m<nmutistate; m++ ) {
		for( n=0; n<nmutistate; n++ ) {
			s=0;
			for( i=0; i<nstate; i++) 
				s += abs(aMutiState[m][i+1]-aMutiState[n][i+1]);
			if (s==2) {	// if muti[m] can become muti[n] by only 1 channel changing state 
				// Find k,l the source and dest states 
				for (i=0; i<nstate; i++) {
					t = aMutiState[m][i+1]-aMutiState[n][i+1];
					if (t==1) 
						k = i;
					if (t==-1) 
						l = i;
					}
				// Check for a path from k to l in model 
				for (i=0; i<npath; i++) {
					if (path[i*4+0]==k && path[i*4+1]==l) {   
						// add a nutipath 
						mutipath[nmutipath*4+0] = m;
						mutipath[nmutipath*4+1] = n;
						mutipath[nmutipath*4+2] = k;
						mutipath[nmutipath*4+3] = l;
						nmutipath++;
						}
					}
				}
			}
		}

	delete [] aMutiState;
	}

//-------------------------------------------------------------------------
// Given data[0..nData-1] and nBin number of bins, calculate bin[o..nBin-1]
// array of bins and idata[0..nData-1] array of bin #'s for data items. 
void skm_binning(int ndata, float* data, int nbin, double* bin/*median*/, int* idata){
	int i;

	//-- Calculate min and max values 
	double nMax=data[0], nMin=data[0];
	for (i=1; i<ndata; i++){
		if(nMax<data[i])
			nMax = data[i];
		if(nMin>data[i])
			nMin = data[i];
		}

	//-- Calculate bins
	double binwidth = (nMax - nMin) / nbin;
	if ( binwidth == 0.0 ) 
		binwidth = 1.0; // avoid divide by 0
	double halfwidth = binwidth / 2.0;
	for ( i=0; i<nbin; i++ )
		bin[i] = nMin + halfwidth + i * binwidth;
	
	//-- Set bin # for each data 
	for ( i=0; i<ndata; i++ )
		idata[i] = max(0, min(nbin-1, int(floor((data[i]-nMin)/binwidth))  ));
	}

//-----------------------------------------------------------------------
// Returns max element in x[0..n-1] and sets imaxpos to position of its 
// first occurrence.
float argmax(int n, float* x, int& imaxpos){
	float xm = x[0];
	imaxpos = 0;
	for( int j=1; j<n; j++) {
		if (xm<x[j]) {
			xm = x[j];
			imaxpos = j;
			}
		}
	
	return xm;
	}	

//------------------------------------------------------------------------
// Viterbi detection
bool skm_viterbi(int ndata, int* data, int nstate, double* pr, double* a,
			 int nbin, double* b, int* istate, double& ll) {
	// data[] : bin index for each data point
	// istate[] : state index for each data point
	// transform probabilities into log domain
	fq::vector<float> logBx( nstate * nbin );
	fq::vector<float> logAx( nstate * nstate );
	fq::vector<float> logPrx( nstate );
	float* logB  = logBx;
	float* logA  = logAx;
	float* logPr = logPrx;
	int i,j;

	for( i=0; i<nstate; i++) {
		for( j=0; j<nbin; j++)       // logB[i][j], b[i][j]
			logB[i*nbin+j] = float( (b[i*nbin+j]>0) ? log(b[i*nbin+j]) : -1.0e10 );
               
		for( j=0; j<nstate; j++)         // logA[i][j], a[i][j]
			logA[i*nstate+j] = float( (a[i*nstate+j]>0) ? log(a[i*nstate+j]) : -1.0e10 );
      
		logPr[i] = float( (pr[i]>0.0) ? log(pr[i]) : -1.0e10 );
		}
	
	// Viterbi recursion
	fq::vector<float> fx(nstate * 2), ux(nstate);
        fq::vector<int> gx(nstate * ndata);
	float *f = fx, *u = ux;
        int *g = gx;
	int fdi; // f data index; alternating 0 or 1

	for( i=0; i<nstate; i++){	
		f[i*2] = logPr[i] + logB[i*nbin+data[0]];	// f[i][0], logB[i][data[0]]
					      //f[i*ndata+0] = logPr[i] + logB[i*nbin+data[0]];   ??
		g[i*ndata+0] = 0;       // g[i][0]
		}
	
	int t;
	for (t=1; t<ndata; t++) {
		for( j=0; j<nstate; j++) {
			fdi = (t - 1) % 2;
			for (int ij=0; ij<nstate; ij++)
				u[ij] = f[ij*2+fdi]+logA[ij*nstate+j]+logB[j*nbin+data[t]];
            //u[i] = f[i*ndata+t-1]+logA[i*nstate+j]+logB[j*nbin+data[t]];
			// f[i][t-1], logA[i][j], logB[j][data[t]]      
			int imax;
			f[j*2+(t%2)] = argmax(nstate,u,imax);  // f[j][t] // sets imax | u[imax] = argmax(...)
			//f[j*ndata+t] = argmax(nstate,u,i);  // f[j][t] // sets imax | u[imax] = argmax(...)
			g[j*ndata+t] = imax;                   // g[j][t]
                        //console << "(" << imax << ", " << f[j*2+(t%2)] << ")\t";
			}
                //console << std::endl;
        }
	
	// backtrack to recover the state sequence 
	for( i=0; i<nstate; i++)
		u[i] = f[i*2+((ndata-1)%2)];             // f[i][ndata-1]
	//u[i] = f[i*ndata+ndata-1];             // f[i][ndata-1]
	int imax;
	ll = argmax(nstate,u,imax);
	istate[ndata-1] = imax;
	
	for (t=ndata-2; t>=0; t--) {
		i = istate[t+1];
		istate[t] = g[i*ndata+t+1];           // g[i][t+1]
		}
	
	return true;
	}

//-----------------------------------------------------------------------
double gaussian(double mean, double xms, double x){
	//const float DELT = float(0.0001);
	const double EPS = 0.00001;
	if ( xms == 0 ) {
		if ( fabs(x-mean) < EPS )
			return 1.0 * SHRT_MAX; // warning this is a sub for infinity
		else
			return 0.0;
	}
	double tmp = (x-mean)/xms;
	tmp = exp(-tmp*tmp/2.0);  // this is where it would underflow
	tmp/= 2.5066283*xms;
	return (tmp<1.0e-8) ? 1.0e-8 : tmp;
	}

//-----------------------------------------------------------------------
// k-means reestimation
void newa(int ndata, int* istate, int nstate, double* a){
	int i, j, s, t;

	// sizeof a is 4 since it is a ptr.  This will set the first member only to 0.
	// TODO - could this possibly be the intention ? 
	// If so, better choice would be sizeof(*a) or sizeof(float)
	// More likely it should be memset(a,0,nState*nState*sizeof(float))
	// Also check if the a array happens to always be 0's on entry to this function.
	// {JB}
	//memset(a, 0, sizeof(a));
	memset(a, 0, nstate*nstate*sizeof(double));
	
	for (t=0; t<ndata-1; t++) {
		i = istate[t];
		j = istate[t+1];
		a[i*nstate+j] += 1.0;                  // a[i][j]
		}
	
	for (i=0; i<nstate; i++) {
		s=0;
		for(j=0; j<nstate; j++) 
			s += int(a[i*nstate+j]);                 // a[i][j]
		if (s > 0.0) {
			for (j=0; j<nstate; j++) 
			  a[i*nstate+j] /= (double) s;              // a[i][j]
			}
		}
	}

//-----------------------------------------------------------------------
void newpr(int nstate, int* istate, double* pr){
	/*
	for (int i=0; i<nstate; i++) 
		pr[i] = (istate[0]==i) ? float(1.0) : float(0.0);
	*/
	for( int i=0; i<nstate; i++) 
		pr[i] = 0.0;
	if(istate[0]>=0 && istate[0]<nstate) //TODO - always true ? -- remove
		pr[istate[0]]=1.0;
	}

//-----------------------------------------------------------------------
void newamp(int ndata, int* idata, int* istate, double* bin, 
			int* state, int nclass, double* amp, int *fixAmp ) {
	double * newamp = new double[nclass];
	int * count = new int[nclass];
	int i, t;
	
	for( i=0; i<nclass; i++) {
		if ( ! fixAmp[i] ) {
			newamp[i] = 0.0;
			count[i] = 0;
			}
		}
  
	for( t=0; t<ndata; t++) {
		i = state[istate[t]];
		if ( ! fixAmp[i] ) {
			newamp[i] += bin[idata[t]];
			count[i] += 1;
			}
		}

	for (i=0; i<nclass; i++)
		if ( count[i] && ! fixAmp[i] )
		  amp[i] = newamp[i] / (double) count[i];
		
	delete [] newamp;
	delete [] count;
	}

//-----------------------------------------------------------------------
void newxms(int ndata, int* idata, int* istate, double* bin,
			int* state, int nclass, double* amp, double* xms, int *fix ) {
	double * newxms = new double[nclass];
	int * count = new int[nclass];
	int i;
	
	for( i=0; i<nclass; i++) {
		if ( ! fix[i] ) {
			newxms[i] = 0.0;
			count[i] = 0;
			}
		}
	
	for( int t=0; t<ndata; t++) {
		i = state[istate[t]];
		if ( ! fix[i] ) {
			newxms[i] += (bin[idata[t]]-amp[i])*(bin[idata[t]]-amp[i]);
			count[i] += 1;
			}
		}
	
	for( i=0; i<nclass; i++)
		if ( count[i] && ! fix[i] )
		  xms[i] = sqrt(newxms[i]/(double)count[i]);
		
	delete [] newxms;
	delete [] count;
	}

//------------------------------------------------------------------------
// skm core routine
int skm_iterate(int ndata, int nbin, double *bin, int *idata, int nstate, int nclass, int* state,
			double* pr, double* a, double* amp, double* xms,  
			int* istate, double& ll, int& it,
			int fFixA, int *fFixI, int *fFixSD, int MaxIter, double tol, 
			int Viterbi, int *Stopped){
    if ( Stopped && *Stopped )
		return -2;
	
	int i;
	
	// skm loop for idealization and reestimation
	fq::vector<double> b( nstate * nbin );
	int fFixPr = 0;
	double  oldll = 0.0f, newll;
        int *test;
	
	for (it=0; it<MaxIter; it++, oldll=ll=newll) {
		for (i=0; i<nstate; i++) 
			for (int j=0; j<nbin; j++)    
				b[i*nbin+j] = gaussian(amp[state[i]],xms[state[i]],bin[j]);
		
		if ( ! skm_viterbi(ndata,idata,nstate,pr,a,nbin,b,istate,newll) )
			return -1;
		
		if ( Stopped && *Stopped )
			return -2;
		
		if ( !Viterbi ) {
		  
			if ( !fFixPr )
				newpr(nstate,istate,pr);
			if ( !fFixA )
				newa(ndata,istate,nstate,a);
			newamp(ndata,idata,istate,bin,state,nclass,amp,fFixI);
			newxms(ndata,idata,istate,bin,state,nclass,amp,xms,fFixSD);
			}
		
		if (it>1 && fabs(newll-oldll)<tol){
			ll = newll;
			break;
			}
		
		if (Viterbi)  {
			ll = newll;
			break;
		}
	}
	return 0;
}


class skm_model {
public:
  int nstate;
  int npath;
  int nclass;
  fq::vector<int> clazz, path;
  fq::vector<double> x, q, amp, xms;
  fq::vector<int> srcCls, destCls;

  skm_model(qubx_model *src)
    : nstate(src->Nstate)//, ...
  {
    // condense the classes--only occupied ones
    int maxcls = -1;
    for (int i=0; i<nstate; ++i) {
      int c = src->clazz[i];
      if ( c > maxcls )
        maxcls = c;
    }
    std::vector<int> occupied(maxcls+1);
    for (int i=0; i<nstate; ++i)
      occupied[src->clazz[i]] = 1;
    destCls.resize(maxcls+1);
    for (int i=0; i<=maxcls; ++i) {
      if ( occupied[i] ) {
        destCls[i] = srcCls.size();
        srcCls.push_back(i);
      }
    }

    for (int i=0; i<nstate; ++i) {
      clazz.push_back(destCls[src->clazz[i]]);
    }

    nclass = srcCls.size();
    for (int i=0; i<nclass; ++i) {
      int c = srcCls[i];
      amp.push_back(src->amp[c]);
      xms.push_back(sqrt(src->var[c]));
    }

    for (int a=0; a<nstate; ++a) {
      for (int b=a+1; b<nstate; ++b) {
        if ( src->K0[a][b] ) {
          path.push_back(a);
          path.push_back(b);
          path.push_back(0);
          path.push_back(0);
          x.push_back( log(src->K0[a][b]) );
          x.push_back( src->K1[a][b] );
          path.push_back(b);
          path.push_back(a);
          path.push_back(0);
          path.push_back(0);
          x.push_back( log(src->K0[b][a]) );
          x.push_back( src->K1[b][a] );
        }
      }
    }
    npath = path.size()/4;
    skm_xtoq(nstate, npath, path, NULL, 0, x, q, 0, NULL, NULL);
  }
};

class skm_metamodel {
public:
  skm_model& model;
  int nmutistate, nmuticlass, nmutipath;
  int *mutistate, *mutipath;
  fq::vector<int>   cmutistate, mutiFixAmp, mutiFixSD;
  fq::vector<double> mutiq, mutia, mutipr, mutiamp, mutixms, ratio, excessNoise;
  double dtMS;

  skm_metamodel(skm_model& _model, int nchannel, double _dtMS)
    : model(_model), mutistate(0), mutipath(0), dtMS(_dtMS)
  {
    ratio.resize(model.nclass);
    // ratio: same as amp but offset so ratio[0] = 0.0; used to classify mutistates
    excessNoise.resize(model.nclass);
    // excessNoise: like ratio for std.dev; used to construct multichannel xms
    double var0 = 0.0;
    if ( model.nclass ) {
      ratio[0] = 0.0;
      excessNoise[0] = 0.0;
      var0 = model.xms[0];
      var0 *= var0;
    }
    
    for (int i=1; i<model.nclass; ++i ) {
      ratio[i] = model.amp[i] - model.amp[0];
      excessNoise[i] = model.xms[i] * model.xms[i] - var0;
    }
    
    skm_mutimakv( nchannel, model.nstate, model.npath, model.clazz, model.path, ratio, 
                  nmutistate, nmutipath, nmuticlass, mutistate, mutipath);
    
    cmutistate.resize( nmutistate );
    for (int i=0; i<nmutistate; ++i )
      cmutistate[i] = mutistate[i*(model.nstate+1)+0]; // mutistate[i][0] -- muticlass of mutistate i
    
    mutiq.resize( nmutistate * nmutistate );
    mutia.resize( nmutistate * nmutistate );
    
    skm_qtomutiq(model.nstate, nmutistate, mutistate, nmutipath, 
                 mutipath, model.q, mutiq, 0, NULL, NULL);

    mutipr.resize( nmutistate );
    
    mutiamp.resize( nmuticlass );
    mutixms.resize( nmuticlass );
    mutiFixAmp.resize( nmuticlass );
    mutiFixSD.resize( nmuticlass );

    reset();
  }
  ~skm_metamodel() {
    delete [] mutistate;
    delete [] mutipath;
  }
  void reset(double _dtMS=0.0) {
    dtMS = _dtMS ? _dtMS : dtMS;
    skm_mexp(nmutistate, float(dtMS*1.0e-3), mutiq, mutia, 0, NULL, NULL, 0); 
    
    for (int i=0; i<nmutistate; ++i )
      mutipr[i] = 1.0 / nmutistate;

    for (int i=0; i<nmuticlass; i++) {
      mutiamp[i] = model.amp[0]; // closed_amp + sum(open_amp === ratio)
      //mutixms[i] = 0.0;          // sqrt( sum( xms*xms ) )
      // now this should probably be // sqrt( closed_noise^2 + sum(excess_noise^2) ) ??
      mutixms[i] = model.xms[0] * model.xms[0]; // = var0; // will be sqrt( closed_noise^2 [var0] + sum(excessNoise^2) )
      //float sum=0.0; // # states contributing to this mutistate; should always be == nchannel
      
      int k;
      int nstate = model.nstate;
      for (k=0; mutistate[k*(nstate+1)+0]!=i; k++)    // mutistate[k][0]
        ; // find the first mutistate in this muticlass.
      // if (e.g. 2 sub-opens == 1 open 1 closed)
      //   then the first example may not have representative s.d. and allXFixed
      
      for (int j=0; j<nstate; j++) {                      // mutistate[k][j+1]
        int countStateJ = mutistate[k*(nstate+1)+j+1];
        mutiamp[i] += countStateJ*ratio[model.clazz[j]];
        mutixms[i] += countStateJ*excessNoise[model.clazz[j]];
        //sum += countStateJ;
      }
      // mutixms[i] /= sum;
      mutixms[i] = sqrt( mutixms[i] );
    }
  }
  void setFix(int *fixAmp, int *fixStd) {
    int nstate = model.nstate;
    for (int i=0; i<nmuticlass; i++) {
      bool allAmpFixed = true;
      bool allSDFixed = true;
      
      int k;
      for (k=0; mutistate[k*(nstate+1)+0]!=i; k++)    // mutistate[k][0]
        ; // find the first mutistate in this muticlass.
      
      for (int j=0; j<nstate; j++) {                      // mutistate[k][j+1]
        int countStateJ = mutistate[k*(nstate+1)+j+1];
        if ( countStateJ &&    // it's a contributor (multiclass i involves one or more channel in class of j)
             (model.clazz[j]      // and either this one is not the baseline
              || (i==0)  ) ) { // or we're in the baseline multiclass (since baseline contributes only to multi-baseline)
          if ( ! (fixAmp && fixAmp[model.srcCls[model.clazz[j]]]) )  // all it takes is one unfixed contributor to make the multiclass unfixed.
            allAmpFixed = false;
          if ( ! (fixStd && fixStd[model.srcCls[model.clazz[j]]]) )
            allSDFixed = false;
        }
      }
      mutiFixAmp[i] = allAmpFixed;
      mutiFixSD[i] = allSDFixed;
    }
  }
};

class skm_session {
public:
  double dtMS;
  qubx_model *model;
  skm_model smodel;
  skm_metamodel mmodel;
  int maxN, N;
  float *data;
  fq::vector<double> bin;
  fq::vector<int> idata, stateSeq;

  int Nstate, Nclass;
  
  skm_session(double _dtMS, int _maxN, qubx_model *_model)
    : dtMS(_dtMS), model(_model), smodel(model), mmodel(smodel, model->Nchannel, _dtMS),
      maxN(_maxN), N(0), stateSeq(maxN)
  {
  }
  void reset(double _dtMS=0.0) {
    dtMS = _dtMS ? _dtMS : dtMS;
    mmodel.reset(dtMS);
  }

  void readStateSeq(int *stSeq) {
    if ( stateSeq && stSeq ) {
      memcpy(stSeq, stateSeq, N*sizeof(int));
    }
  }

  void useData(int n, float *_data) {
    N = n;
    data = _data;
    idata.resize(n);
    stateSeq.resize(n);
    
    int nbin  = 100;
    // we compromise between std.dev resolution and number of bins
    double minxms = mmodel.mutixms[0];
    for (int i=1; i<mmodel.nmuticlass; ++i)
      minxms = min(minxms, mmodel.mutixms[i]);
    if ( minxms == 0.0 ) minxms = 1e-3;
    nbin = max(10, min(10240, (int) (20 * fabs(mmodel.mutiamp[mmodel.nmuticlass-1] - mmodel.mutiamp[0]) / minxms)));
    minxms = 5 * fabs(mmodel.mutiamp[mmodel.nmuticlass-1] - mmodel.mutiamp[0]) / nbin;
    if ( minxms == 0.0 ) minxms = 1e-3;
    for (int i=0; i<mmodel.nmuticlass; ++i)
      mmodel.mutixms[i] = max(minxms, mmodel.mutixms[i]);
    
    bin.resize(nbin);
    skm_binning(n,_data,nbin,bin,idata);
  }
  
  void viterbi(int n, float *_data, int *stSeq, double *ll) {
    reset(); // mutixms could be modified by binning
    useData(n, _data);
    /*
    console << n << " points" << std::endl;
    for (int i=0; i<12; ++i)
      console << "\t" << _data[n-1-i] << "\t" << idata[n-1-i] << std::endl;
    for (int i=0; i<mmodel.nmuticlass; ++i)
      console << mmodel.mutiamp[i] << "\t" << mmodel.mutixms[i] << std::endl;
    for (int i=0; i<mmodel.nmutistate; ++i) {
      for (int j=0; j<mmodel.nmutistate; ++j)
        console << mmodel.mutia[i*mmodel.nmutistate+j] << "\t";
      console << std::endl;
    }
    */
    int iter;
    int err = skm_iterate(n, bin.size(), bin, idata, mmodel.nmutistate, mmodel.nmuticlass, mmodel.cmutistate,
                          mmodel.mutipr, mmodel.mutia, mmodel.mutiamp, mmodel.mutixms,
                          stateSeq, *ll, iter, 1, mmodel.mutiFixAmp, mmodel.mutiFixSD, 1, SKM_LLCONV, 1, 0);
    if ( err ) {
      console << "skm error: " << err << endl;
    }
    readStateSeq(stSeq);
  }

  int iterate(int n, float *_data, int *stSeq, int maxIter, int fixA, int *fixAmp, int *fixStd, double *ll) {
    reset();
    useData(n, _data);
    mmodel.setFix(fixAmp, fixStd);
    /*
    console << n << " points, " << maxIter << " iterations, " << fixA << ", " << fixAmp[0] << ", " << fixStd[0] << ", " << fixAmp[1] << ", " << fixStd[1] << std::endl;
    for (int i=0; i<12; ++i)
      console << "\t" << _data[n-1-i] << "\t" << idata[n-1-i] << std::endl;
    for (int i=0; i<mmodel.nmuticlass; ++i)
      console << mmodel.mutiamp[i] << "\t" << mmodel.mutixms[i] << std::endl;
    for (int i=0; i<mmodel.nmutistate; ++i) {
      for (int j=0; j<mmodel.nmutistate; ++j)
        console << mmodel.mutia[i*mmodel.nmutistate+j] << "\t";
      console << std::endl;
    }
    */
    int iter = 0;
    int err = skm_iterate(n, bin.size(), bin, idata, mmodel.nmutistate, mmodel.nmuticlass, mmodel.cmutistate,
                          mmodel.mutipr, mmodel.mutia, mmodel.mutiamp, mmodel.mutixms,
                          stateSeq, *ll, iter, fixA, mmodel.mutiFixAmp, mmodel.mutiFixSD,
                          maxIter, SKM_LLCONV, 0, 0);
    if ( err ) {
      console << "skm error: " << err << endl;
    }
    readStateSeq(stSeq);
    return iter;
  }

  int countEvents() {
    if ( N == 0 ) {
      return 0;
    }
    int *clazz = mmodel.cmutistate;
    int *ss = stateSeq;
    int c = clazz[ss[0]];
    int Nevent = 1;
    for (int i=1; i<N; ++i) {
      int d = clazz[ss[i]];
      if ( d != c ) {
        ++Nevent;
        c = d;
      }
    }
    //console << Nevent << " events" << std::endl;
    return Nevent;
  }

  void readEvents(int *classes, int *counts) {
    if ( N == 0 ) {
      return;
    }
    int *clazz = mmodel.cmutistate;
    int *ss = stateSeq;
    int c = clazz[ss[0]];
    int count = 1;
    int iOut = 0;
    for (int i=1; i<N; ++i) {
      int d = clazz[ss[i]];
      if ( d != c ) {
        classes[iOut] = c;
        counts[iOut] = count;
        ++iOut;
        c = d;
        count = 1;
      } else {
        ++count;
      }
    }
    classes[iOut] = c;
    counts[iOut] = count;
    ++iOut;

    if ( model->Nchannel == 1 ) {
      for (int i=0; i<iOut; ++i) {
        classes[i] = smodel.srcCls[classes[i]];
      }
    }
  }
};
  
extern "C" void* QUBFAST_API qubx_skm_session_create(double dtMS, int maxN, qubx_model *model) {
  return new skm_session(dtMS, maxN, model);
}
extern "C" void QUBFAST_API qubx_skm_session_free(void *skm) {
  delete (skm_session *) skm;
}
extern "C" void QUBFAST_API qubx_skm_session_reset(void *skm, double dtMS) {
  skm_session *s = (skm_session *) skm;
  s->reset(dtMS);
}

extern "C" void QUBFAST_API qubx_skm_viterbi(void *skm, int n, float *data, int *stateSeq, double *ll) {
  skm_session *s = (skm_session *) skm;
  s->viterbi(n, data, stateSeq, ll);
}
  
extern "C" int QUBFAST_API qubx_skm_iterate(void *skm, int n, float *data, int *stateSeq, int maxIter, int fixA, int *fixAmp, int *fixStd, double *ll) {
  skm_session *s = (skm_session *) skm;
  return s->iterate(n, data, stateSeq, maxIter, fixA, fixAmp, fixStd, ll);
}

extern "C" int QUBFAST_API qubx_skm_Nstate(void *skm) {
  skm_session *s = (skm_session *) skm;
  return s->mmodel.nmutistate;
}

extern "C" void QUBFAST_API qubx_skm_read_A(void *skm, double *A) {
  skm_session *s = (skm_session *) skm;
  int n = s->mmodel.nmutistate;
  memcpy(A, s->mmodel.mutia, n*n*sizeof(double));
}
extern "C" int QUBFAST_API qubx_skm_Nclass(void *skm) {
  skm_session *s = (skm_session *) skm;
  return s->mmodel.nmuticlass;
}
extern "C" void QUBFAST_API qubx_skm_read_amp(void *skm, double *amp, double *std) {
  skm_session *s = (skm_session *) skm;
  int n = s->mmodel.nmuticlass;
  memcpy(amp, s->mmodel.mutiamp, n*sizeof(double));
  memcpy(std, s->mmodel.mutixms, n*sizeof(double));
}
extern "C" int QUBFAST_API qubx_skm_Nevent(void *skm) {
  skm_session *s = (skm_session *) skm;
  return s->countEvents();
}
extern "C" void QUBFAST_API qubx_skm_read_events(void *skm, int *classes, int *counts) {
  skm_session *s = (skm_session *) skm;
  s->readEvents(classes, counts);
}
